@@ -30,7 +30,7 @@
 #include "cap.mdh"
 #include "cap.pro"
 
-#ifdef HAVE_CAP_INIT
+#ifdef HAVE_CAP_GET_PROC
 
 static int
 bin_cap(char *nam, char **argv, char *ops, int func)
@@ -106,13 +106,13 @@ bin_setcap(char *nam, char **argv, char *ops, int func)
     return ret;
 }
 
-#else /* !HAVE_CAP_INIT */
+#else /* !HAVE_CAP_GET_PROC */
 
 # define bin_cap    bin_notavail
 # define bin_getcap bin_notavail
 # define bin_setcap bin_notavail
 
-#endif /* !HAVE_CAP_INIT */
+#endif /* !HAVE_CAP_GET_PROC */
 
 /* module paraphernalia */
 
