@@ -442,7 +442,13 @@ digitargument(void)
 
     if (!(zmod.flags & MOD_TMULT))
 	zmod.tmult = 0;
-    zmod.tmult = zmod.tmult * 10 + sign * (c & 0xf);
+    if (zmod.flags & MOD_NEG) {
+	/* If we just had a negative argument, this is the digit, *
+	 * rather than the -1 assumed by negargument()            */
+	zmod.tmult = sign * (c & 0xf);
+	zmod.flags &= ~MOD_NEG;
+    } else
+	zmod.tmult = zmod.tmult * 10 + sign * (c & 0xf);
     zmod.flags |= MOD_TMULT;
     prefixflag = 1;
 }
@@ -456,7 +462,7 @@ negargument(void)
 	return;
     }
     zmod.tmult = -1;
-    zmod.flags |= MOD_TMULT;
+    zmod.flags |= MOD_TMULT|MOD_NEG;
     prefixflag = 1;
 }
 
@@ -464,7 +470,23 @@ negargument(void)
 void
 universalargument(void)
 {
-    zmod.tmult *= 4;
+    int digcnt = 0, pref = 0, minus = 1, gotk;
+    while ((gotk = getkey(0)) != EOF) {
+	if (gotk == '-' && !digcnt) {
+	    minus = -1;
+	    digcnt++;
+	} else if (gotk >= '0' && gotk <= '9') {
+	    pref = pref * 10 + (gotk & 0xf);
+	    digcnt++;
+	} else {
+	    ungetkey(gotk);
+	    break;
+	}
+    }
+    if (digcnt)
+	zmod.tmult = minus * (pref ? pref : 1);
+    else
+	zmod.tmult *= 4;
     zmod.flags |= MOD_TMULT;
     prefixflag = 1;
 }
