@@ -468,7 +468,9 @@ bin_zle_call(char *name, char **args, char *ops, char func)
 	return 1;
     }
     t = rthingy(args[0]);
-    execzlefunc(t);
+    PERMALLOC {
+      execzlefunc(t);
+    } LASTALLOC;
     unrefthingy(t);
     return 0;
 }
