@@ -84,7 +84,7 @@ spaceinline(int ct)
     ll += ct;
     line[ll] = '\0';
 
-    if (mark >= cs)
+    if (mark > cs)
 	mark += ct;
 }
 
@@ -516,6 +516,7 @@ mkundoent(void)
 	line[ll - 1 - suf] == lastline[lastll - 1 - suf]; )
 	suf++;
     ch = zalloc(sizeof(*ch));
+    ch->next = NULL;
     ch->hist = histline;
     ch->off = pre;
     if(suf + pre == lastll)
