@@ -106,7 +106,7 @@ static struct builtin builtins[] =
     BUILTIN("times", BINF_PSPECIAL, bin_times, 0, 0, 0, NULL, NULL),
     BUILTIN("trap", BINF_PSPECIAL, bin_trap, 0, -1, 0, NULL, NULL),
     BUILTIN("true", 0, bin_true, 0, -1, 0, NULL, NULL),
-    BUILTIN("type", 0, bin_whence, 0, -1, 0, "ampfs", "v"),
+    BUILTIN("type", 0, bin_whence, 0, -1, 0, "ampfsw", "v"),
     BUILTIN("typeset", BINF_TYPEOPTS | BINF_MAGICEQUALS | BINF_PSPECIAL, bin_typeset, 0, -1, 0, "LRUZfilrtuxm", NULL),
     BUILTIN("umask", 0, bin_umask, 0, 1, 0, "S", NULL),
     BUILTIN("unalias", 0, bin_unhash, 1, -1, 0, "m", "a"),
@@ -115,9 +115,9 @@ static struct builtin builtins[] =
     BUILTIN("unset", BINF_PSPECIAL, bin_unset, 1, -1, 0, "fm", NULL),
     BUILTIN("unsetopt", 0, bin_setopt, 0, -1, BIN_UNSETOPT, NULL, NULL),
     BUILTIN("wait", 0, bin_fg, 0, -1, BIN_WAIT, NULL, NULL),
-    BUILTIN("whence", 0, bin_whence, 0, -1, 0, "acmpvfs", NULL),
-    BUILTIN("where", 0, bin_whence, 0, -1, 0, "pms", "ca"),
-    BUILTIN("which", 0, bin_whence, 0, -1, 0, "amps", "c"),
+    BUILTIN("whence", 0, bin_whence, 0, -1, 0, "acmpvfsw", NULL),
+    BUILTIN("where", 0, bin_whence, 0, -1, 0, "pmsw", "ca"),
+    BUILTIN("which", 0, bin_whence, 0, -1, 0, "ampsw", "c"),
 
 #ifdef DYNAMIC
     BUILTIN("zmodload", 0, bin_zmodload, 0, -1, 0, "Laudi", NULL),
@@ -162,6 +162,11 @@ printbuiltinnode(HashNode hn, int printflags)
 {
     Builtin bn = (Builtin) hn;
 
+    if (printflags & PRINT_WHENCE_WORD) {
+	printf("%s: builtin\n", bn->nam);
+	return;
+    }
+
     if (printflags & PRINT_WHENCE_CSH) {
 	printf("%s: shell built-in command\n", bn->nam);
 	return;
@@ -607,6 +612,39 @@ bin_dirs(char *name, char **argv, char *ops, int func)
 
 /* cd, chdir, pushd, popd */
 
+/**/
+void
+set_pwd_env(void)
+{
+    Param pm;
+
+    pm = (Param) paramtab->getnode(paramtab, "PWD");
+    if (pm && PM_TYPE(pm->flags) != PM_SCALAR) {
+	pm->flags &= ~PM_READONLY;
+	unsetparam_pm(pm, 0, 1);
+    }
+
+    pm = (Param) paramtab->getnode(paramtab, "OLDPWD");
+    if (pm && PM_TYPE(pm->flags) != PM_SCALAR) {
+	pm->flags &= ~PM_READONLY;
+	unsetparam_pm(pm, 0, 1);
+    }
+
+    setsparam("PWD", ztrdup(pwd));
+    setsparam("OLDPWD", ztrdup(oldpwd));
+
+    pm = (Param) paramtab->getnode(paramtab, "PWD");
+    if (!(pm->flags & PM_EXPORTED)) {
+	pm->flags |= PM_EXPORTED;
+	pm->env = addenv("PWD", pwd);
+    }
+    pm = (Param) paramtab->getnode(paramtab, "OLDPWD");
+    if (!(pm->flags & PM_EXPORTED)) {
+	pm->flags |= PM_EXPORTED;
+	pm->env = addenv("PWD", pwd);
+    }
+}
+
 /* The main pwd changing function.  The real work is done by other     *
  * functions.  cd_get_dest() does the initial argument processing;     *
  * cd_do_chdir() actually changes directory, if possible; cd_new_pwd() *
@@ -666,6 +704,7 @@ bin_cd(char *nam, char **argv, char *ops, int func)
 	    chdir(unmeta(pwd));
 	}
     }
+    set_pwd_env();
     return 0;
 }
 
@@ -1586,8 +1625,8 @@ bin_typeset(char *name, char **argv, char *ops, int func)
 		pm->ct = auxlen;
 	    if (PM_TYPE(pm->flags) != PM_ARRAY) {
 		if (pm->flags & PM_EXPORTED) {
-		    if (!pm->env)
-			pm->env = addenv(asg->name, (asg->value) ? asg->value : getsparam(asg->name));
+		    if (!(pm->flags & PM_UNSET) && !pm->env && !asg->value)
+			pm->env = addenv(asg->name, getsparam(asg->name));
 		} else if (pm->env) {
 		    delenv(pm->env);
 		    zsfree(pm->env);
@@ -1807,7 +1846,7 @@ bin_whence(char *nam, char **argv, char *ops, int func)
     Comp com;
     int returnval = 0;
     int printflags = 0;
-    int csh, all, v;
+    int csh, all, v, wd;
     int informed;
     char *cnam;
 
@@ -1815,8 +1854,11 @@ bin_whence(char *nam, char **argv, char *ops, int func)
     csh = ops['c'];
     v   = ops['v'];
     all = ops['a'];
+    wd  = ops['w'];
 
-    if (ops['c'])
+    if (ops['w'])
+	printflags |= PRINT_WHENCE_WORD;
+    else if (ops['c'])
 	printflags |= PRINT_WHENCE_CSH;
     else if (ops['v'])
 	printflags |= PRINT_WHENCE_VERBOSE;
@@ -1919,33 +1961,41 @@ bin_whence(char *nam, char **argv, char *ops, int func)
 		    continue;
 		strcpy(z, *argv);
 		if (iscom(buf)) {
-		    if (v && !csh)
-			zputs(*argv, stdout), fputs(" is ", stdout);
-		    zputs(buf, stdout);
-		    if (ops['s'])
-			print_if_link(buf);
-		    fputc('\n', stdout);
+		    if (wd) {
+			printf("%s: command\n", *argv);
+		    } else {
+			if (v && !csh)
+			    zputs(*argv, stdout), fputs(" is ", stdout);
+			zputs(buf, stdout);
+			if (ops['s'])
+			    print_if_link(buf);
+			fputc('\n', stdout);
+		    }
 		    informed = 1;
 		}
 	    }
-	    if (!informed && (v || csh)) {
+	    if (!informed && (wd || v || csh)) {
 		zputs(*argv, stdout);
-		puts(" not found");
+		puts(wd ? ": none" : " not found");
 		returnval = 1;
 	    }
 	} else if ((cnam = findcmd(*argv))) {
 	    /* Found external command. */
-	    if (v && !csh)
-		zputs(*argv, stdout), fputs(" is ", stdout);
-	    zputs(cnam, stdout);
-	    if (ops['s'])
-		print_if_link(cnam);
-	    fputc('\n', stdout);
+	    if (wd) {
+		printf("%s: command\n", *argv);
+	    } else {
+		if (v && !csh)
+		    zputs(*argv, stdout), fputs(" is ", stdout);
+		zputs(cnam, stdout);
+		if (ops['s'])
+		    print_if_link(cnam);
+		fputc('\n', stdout);
+	    }
 	    zsfree(cnam);
 	} else {
 	    /* Not found at all. */
-	    if (v || csh)
-		zputs(*argv, stdout), puts(" not found");
+	    if (v || csh || wd)
+		zputs(*argv, stdout), puts(wd ? ": none" : " not found");
 	    returnval = 1;
 	}
     }
@@ -2425,12 +2475,12 @@ bin_echotc(char *name, char **argv, char *ops, int func)
     /* if the specified termcap is boolean, and set, say so  *
      * ncurses can tell if an existing boolean capability is *
      * off so in this case we print "no".                    */
-#ifndef NCURSES_VERSION
+#if !defined(NCURSES_VERSION) || !defined(COLOR_PAIR)
     if (tgetflag(s) > 0) {
 	puts("yes");
 	return (0);
     }
-#else
+#else /* NCURSES_VERSION && COLOR_PAIR */
     switch (tgetflag(s)) {
     case -1:
 	break;
@@ -2441,7 +2491,7 @@ bin_echotc(char *name, char **argv, char *ops, int func)
 	puts("yes");
 	return 0;
     }
-#endif
+#endif /* NCURSES_VERSION && COLOR_PAIR */
     /* get a string-type capability */
     u = buf;
     t = tgetstr(s, &u);
@@ -2530,7 +2580,7 @@ bin_getopts(char *name, char **argv, char *ops, int func)
     char *optstr = unmetafy(*argv++, &lenoptstr), *var = *argv++;
     char **args = (*argv) ? argv : pparams;
     static int optcind = 0;
-    char *str, optbuf[2], *p, opch;
+    char *str, optbuf[2] = " ", *p, opch;
 
     /* zoptind keeps count of the current argument number.  The *
      * user can set it to zero to start a new option parse.     */
@@ -2551,10 +2601,10 @@ bin_getopts(char *name, char **argv, char *ops, int func)
     /* find place in relevant argument */
     str = unmetafy(dupstring(args[zoptind - 1]), &lenstr);
     if(optcind >= lenstr) {
+	optcind = 0;
 	if(!args[zoptind++])
 	    return 1;
 	str = unmetafy(dupstring(args[zoptind - 1]), &lenstr);
-	optcind = 0;
     }
     if(!optcind) {
 	if(lenstr < 2 || (*str != '-' && *str != '+'))
@@ -2577,13 +2627,14 @@ bin_getopts(char *name, char **argv, char *ops, int func)
     if(opch == ':' || !(p = memchr(optstr, opch, lenoptstr))) {
 	p = "?";
 err:
+      zsfree(zoptarg);
 	if(quiet) {
 	    setsparam(var, ztrdup(p));
-	    zsfree(zoptarg);
 	    zoptarg = metafy(optbuf, lenoptbuf, META_DUP);
 	} else {
 	    zerr(*p == '?' ? "bad option: -%c" :
 		"argument expected after -%c option", NULL, opch);
+          zoptarg=ztrdup("");
 	    errflag = 0;
 	}
 	return 0;
@@ -3176,7 +3227,7 @@ bin_read(char *name, char **args, char *ops, int func)
 static int
 zread(void)
 {
-    char cc;
+    char cc, retry = 0;
 
     /* use zbuf if possible */
     if (zbuf)
@@ -3187,12 +3238,33 @@ zread(void)
 	    return zbuf++, STOUC(*zbuf++ ^ 32);
 	else
 	    return (*zbuf) ? STOUC(*zbuf++) : EOF;
-    /* read a character from readfd */
-    if (read(readfd, &cc, 1) != 1)
-	/* on EOF, return EOF */
+    for (;;) {
+	/* read a character from readfd */
+	switch (read(readfd, &cc, 1)) {
+	case 1:
+	    /* return the character read */
+	    return STOUC(cc);
+#if defined(EAGAIN) || defined(EWOULDBLOCK)
+	case -1:
+	    if (!retry && readfd == 0 && (
+# ifdef EAGAIN
+		    errno == EAGAIN
+#  ifdef EWOULDBLOCK
+		    ||
+#  endif /* EWOULDBLOCK */
+# endif /* EAGAIN */
+# ifdef EWOULDBLOCK
+		    errno == EWOULDBLOCK
+# endif /* EWOULDBLOCK */
+		) && setblock_stdin()) {
+		retry = 1;
+		continue;
+	    }
+	    break;
+#endif /* EAGAIN || EWOULDBLOCK */
+	}
 	return EOF;
-    /* return the character read */
-    return STOUC(cc);
+    }
 }
 
 /* holds arguments for testlex() */
