@@ -53,12 +53,16 @@ strstr(const char *s, const char *t)
 
 #ifndef HAVE_GETHOSTNAME
 int
-gethostname(char *name, int namelen)
+gethostname(char *name, size_t namelen)
 {
-    struct utsname uname_str;
+    struct utsname uts;
 
-    uname(&uname_str);
-    strncpy(hostnam, uname_str.nodename, 256);
+    uname(&uts);
+    if(strlen(uts.nodename) >= namelen) {
+	errno = EINVAL;
+	return -1;
+    }
+    strcpy(name, uts.nodename);
     return 0;
 }
 #endif
@@ -169,12 +173,22 @@ zgetdir(struct dirsav *d)
 		(fn[1] == '\0' ||
 		 (fn[1] == '.' && fn[2] == '\0')))
 		continue;
-	    if (dev != pdev || (ino_t) de->d_ino == ino) {
+#ifdef HAVE_STRUCT_DIRENT_D_STAT
+	    if(de->d_stat.st_dev == dev && de->d_stat.st_ino == ino) {
+		strncpy(nbuf + 3, fn, PATH_MAX);
+		break;
+	    }
+#else /* !HAVE_STRUCT_DIRENT_D_STAT */
+# ifdef HAVE_STRUCT_DIRENT_D_INO
+	    if (dev != pdev || (ino_t) de->d_ino == ino)
+# endif /* HAVE_STRUCT_DIRENT_D_INO */
+	    {
 		strncpy(nbuf + 3, fn, PATH_MAX);
 		lstat(nbuf, &sbuf);
 		if (sbuf.st_dev == dev && sbuf.st_ino == ino)
 		    break;
 	    }
+#endif /* !HAVE_STRUCT_DIRENT_D_STAT */
 	}
 	closedir(dir);
 	if (!de)
