@@ -522,6 +522,18 @@ iscom(char *s)
 	    S_ISREG(statbuf.st_mode));
 }
 
+/**/
+int
+isreallycom(Cmdnam cn)
+{
+    char fullnam[MAXCMDLEN];
+
+    strcpy(fullnam, cn->u.name ? *(cn->u.name) : "");
+    strcat(fullnam, "/");
+    strcat(fullnam, cn->nam);
+    return iscom(fullnam);
+}
+
 /**/
 int
 isrelative(char *s)
@@ -710,6 +722,7 @@ execpline(Sublist l, int how, int last1)
 {
     int ipipe[2], opipe[2];
     int pj, newjob;
+    int old_simple_pline = simple_pline;
     static int lastwj;
 
     if (!l->left)
@@ -874,7 +887,7 @@ execpline(Sublist l, int how, int last1)
 	    lastval = !lastval;
     }
     if (!pline_level)
-	simple_pline = 0;
+	simple_pline = old_simple_pline;
     return lastval;
 }
 
@@ -1482,6 +1495,10 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 
     if (type == SIMPLE && !nullexec) {
 	char *s;
+	char trycd = (isset(AUTOCD) && isset(SHINSTDIN)
+		      && empty(cmd->redir) && !empty(args)
+		      && !nextnode(firstnode(args))
+		      && *(char *)peekfirst(args));
 
 	DPUTS(empty(args), "BUG: empty(args) in exec.c");
 	if (!hn) {
@@ -1489,6 +1506,11 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 	    char *cmdarg = (char *) peekfirst(args);
 
 	    hn = cmdnamtab->getnode(cmdnamtab, cmdarg);
+	    if (hn && trycd && !isreallycom((Cmdnam)hn)) {
+		cmdnamtab->removenode(cmdnamtab, cmdarg);
+		cmdnamtab->freenode(hn);
+		hn = NULL;
+	    }
 	    if (!hn && isset(HASHCMDS) && strcmp(cmdarg, "..")) {
 		for (s = cmdarg; *s && *s != '/'; s++);
 		if (!*s)
@@ -1498,9 +1520,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 
 	/* If no command found yet, see if it  *
 	 * is a directory we should AUTOCD to. */
-	if (!hn && isset(AUTOCD) && isset(SHINSTDIN) && empty(cmd->redir)
-	    && !nextnode(firstnode(args)) && *(char *)peekfirst(args)
-	    && (s = cancd(peekfirst(args)))) {
+	if (!hn && trycd && (s = cancd(peekfirst(args)))) {
 	    peekfirst(args) = (void *) s;
 	    pushnode(args, dupstring("cd"));
 	    if ((hn = builtintab->getnode(builtintab, "cd")))
@@ -1575,9 +1595,11 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 	forked = 1;
 	if (sigtrapped[SIGINT] & ZSIG_IGNORED)
 	    holdintr();
+#ifdef HAVE_NICE
 	/* Check if we should run background jobs at a lower priority. */
 	if ((how & Z_ASYNC) && isset(BGNICE))
 	    nice(5);
+#endif /* HAVE_NICE */
 
     } else if (is_cursh) {
 	/* This is a current shell procedure that didn't need to fork.    *
@@ -2242,7 +2264,12 @@ readoutput(int in, int qt)
     fin = fdopen(in, "r");
     ret = newlinklist();
     ptr = buf = (char *) ncalloc(bsiz = 64);
-    while ((c = fgetc(fin)) != EOF) {
+    while ((c = fgetc(fin)) != EOF || errno == EINTR) {
+	if (c == EOF) {
+	    errno = 0;
+	    clearerr(fin);
+	    continue;
+	}
 	if (imeta(c)) {
 	    *ptr++ = Meta;
 	    c ^= 32;
