@@ -630,9 +630,23 @@ parsecomp(int gflag)
 		return NULL;
 	} else if (*pptr == Inbrack) {
 	    /* Character set: brackets had better match */
-	    while (*++pptr && *pptr != Outbrack)
-		if (itok(*pptr))
+	    if (pptr[1] == Outbrack)
+		*++pptr = ']';
+	    else if ((pptr[1] == Hat || pptr[1] == '^' || pptr[1] == '!') &&
+		     pptr[2] == Outbrack)
+		*(pptr += 2) = ']';
+	    while (*++pptr && *pptr != Outbrack) {
+		if (itok(*pptr)) {
+		    /* POSIX classes: make sure it's a real one, *
+		     * leave the Inbrack tokenised if so.        */
+		    char *nptr;
+		    if (*pptr == Inbrack && pptr[1] == ':'
+			&& (nptr = strchr(pptr+2, ':')) && 
+			*++nptr == Outbrack)
+			pptr = nptr;
 		    *pptr = ztokens[*pptr - Pound];
+		}
+	    }
 	    if (*pptr != Outbrack)
 		return NULL;
 	} else if (itok(*pptr) && *pptr != Star && *pptr != Quest)
@@ -1050,7 +1064,6 @@ glob(LinkList list, LinkNode np)
 			    data = qgetnum(&s);
 			else {
 			    /* ... or a user name */
-			    struct passwd *pw;
 			    char sav, *tt;
 
 			    /* Find matching delimiters */
@@ -1060,6 +1073,8 @@ glob(LinkList list, LinkNode np)
 				     NULL, 0);
 				data = 0;
 			    } else {
+#ifdef HAVE_GETPWNAM
+				struct passwd *pw;
 				sav = *tt;
 				*tt = '\0';
 
@@ -1069,7 +1084,13 @@ glob(LinkList list, LinkNode np)
 				    zerr("unknown user", NULL, 0);
 				    data = 0;
 				}
-				if ((*tt = sav))
+				*tt = sav;
+#else /* !HAVE_GETPWNAM */
+				sav = *tt;
+				zerr("unknown user", NULL, 0);
+				data = 0;
+#endif /* !HAVE_GETPWNAM */
+				if (sav)
 				    s = tt + 1;
 				else
 				    s = tt;
@@ -1084,7 +1105,6 @@ glob(LinkList list, LinkNode np)
 			    data = qgetnum(&s);
 			else {
 			    /* ...or a delimited group name. */
-			    struct group *gr;
 			    char sav, *tt;
 
 			    tt = get_strarg(s);
@@ -1093,6 +1113,8 @@ glob(LinkList list, LinkNode np)
 				     NULL, 0);
 				data = 0;
 			    } else {
+#ifdef HAVE_GETGRNAM
+				struct group *gr;
 				sav = *tt;
 				*tt = '\0';
 
@@ -1102,7 +1124,13 @@ glob(LinkList list, LinkNode np)
 				    zerr("unknown group", NULL, 0);
 				    data = 0;
 				}
-				if ((*tt = sav))
+				*tt = sav;
+#else /* !HAVE_GETGRNAM */
+				sav = *tt;
+				zerr("unknown group", NULL, 0);
+				data = 0;
+#endif /* !HAVE_GETGRNAM */
+				if (sav)
 				    s = tt + 1;
 				else
 				    s = tt;
@@ -2187,6 +2215,90 @@ doesmatch(Comp c)
 	return matchonce(c);
 }
 
+/**/
+static int
+posix_range(char **patptr, int ch)
+{
+    /* Match POSIX ranges, which correspond to ctype macros,  *
+     * e.g. [:alpha:] -> isalpha.  It just doesn't seem worth * 
+     * the palaver of creating a hash table for this.         */
+    char *start = *patptr;
+    int len;
+
+    /* we made sure in parsecomp() there was a ':' to search for */
+    *patptr = strchr(start, ':');
+    len = (*patptr)++ - start;
+
+    if (!strncmp(start, "alpha", len))
+	return isalpha(ch);
+    if (!strncmp(start, "alnum", len))
+	return isalnum(ch);
+    if (!strncmp(start, "blank", len))
+	return ch == ' ' || ch == '\t';
+    if (!strncmp(start, "cntrl", len))
+	return iscntrl(ch);
+    if (!strncmp(start, "digit", len))
+	return isdigit(ch);
+    if (!strncmp(start, "graph", len))
+	return isgraph(ch);
+    if (!strncmp(start, "lower", len))
+	return islower(ch);
+    if (!strncmp(start, "print", len))
+	return isprint(ch);
+    if (!strncmp(start, "punct", len))
+	return ispunct(ch);
+    if (!strncmp(start, "space", len))
+	return isspace(ch);
+    if (!strncmp(start, "upper", len))
+	return isupper(ch);
+    if (!strncmp(start, "xdigit", len))
+	return isxdigit(ch);
+    return 0;
+}
+
+/**/
+static void
+rangematch(char **patptr, int ch, int rchar)
+{
+    /* Check for a character in a [...] or [^...].  The [ *
+     * and optional ^ have already been skipped.          */
+
+    char *pat = *patptr;
+#ifdef HAVE_STRCOLL
+    char l_buf[2], r_buf[2], ch_buf[2];
+
+    ch_buf[0] = ch;
+    l_buf[1] = r_buf[1] = ch_buf[1] = '\0';
+#endif
+
+#define PAT(X) (pat[X] == Meta ? pat[(X)+1] ^ 32 : untok(pat[X]))
+#define PPAT(X) (pat[(X)-1] == Meta ? pat[X] ^ 32 : untok(pat[X]))
+
+    for (pat++; *pat != Outbrack && *pat;
+	 *pat == Meta ? pat += 2 : pat++) {
+	if (*pat == Inbrack) {
+	    /* Inbrack can only occur inside a range if we found [:...:]. */
+	    pat += 2;
+	    if (posix_range(&pat, ch))
+		break;
+	} else if (*pat == '-' && pat[-1] != rchar &&
+		   pat[1] != Outbrack) {
+#ifdef HAVE_STRCOLL
+	    l_buf[0] = PPAT(-1);
+	    r_buf[0] = PAT(1);
+	    if (strcoll(l_buf, ch_buf) <= 0 &&
+		strcoll(ch_buf, r_buf) <= 0)
+#else
+		if (PPAT(-1) <= ch && PAT(1) >= ch)
+#endif
+		    break;
+	} else if (ch == PAT(0))
+	    break;
+    }
+
+    *patptr = pat;
+}
+
 /**/
 static int
 matchonce(Comp c)
@@ -2301,38 +2413,15 @@ matchonce(Comp c)
 	}
 	if (*pat == Inbrack) {
 	    /* Match groups of characters */
-#define PAT(X) (pat[X] == Meta ? pat[(X)+1] ^ 32 : untok(pat[X]))
-#define PPAT(X) (pat[(X)-1] == Meta ? pat[X] ^ 32 : untok(pat[X]))
 	    char ch;
-#ifdef HAVE_STRCOLL
-	    char l_buf[2], r_buf[2], ch_buf[2];
-
-	    l_buf[1] = r_buf[1] = ch_buf[1] = '\0';
-#endif
 
 	    if (!*pptr)
 		break;
 	    ch = *pptr == Meta ? pptr[1] ^ 32 : *pptr;
-#ifdef HAVE_STRCOLL
-	    ch_buf[0] = ch;
-#endif
 	    if (pat[1] == Hat || pat[1] == '^' || pat[1] == '!') {
 		/* group is negated */
-		pat[1] = Hat;
-		for (pat += 2; *pat != Outbrack && *pat;
-		     *pat == Meta ? pat += 2 : pat++)
-		    if (*pat == '-' && pat[-1] != Hat && pat[1] != Outbrack) {
-#ifdef HAVE_STRCOLL
-			l_buf[0] = PPAT(-1);
-			r_buf[0] = PAT(1);
-			if (strcoll(l_buf, ch_buf) <= 0 &&
-			    strcoll(ch_buf, r_buf) <= 0)
-#else
-			if (PPAT(-1) <= ch && PAT(1) >= ch)
-#endif
-			    break;
-		    } else if (ch == PAT(0))
-			break;
+		*++pat = Hat;
+		rangematch(&pat, ch, Hat);
 		DPUTS(!*pat, "BUG: something is very wrong in doesmatch()");
 		if (*pat != Outbrack)
 		    break;
@@ -2341,21 +2430,7 @@ matchonce(Comp c)
 		continue;
 	    } else {
 		/* pattern is not negated (affirmed? asserted?) */
-		for (pat++; *pat != Outbrack && *pat;
-		     *pat == Meta ? pat += 2 : pat++)
-		    if (*pat == '-' && pat[-1] != Inbrack &&
-			       pat[1] != Outbrack) {
-#ifdef HAVE_STRCOLL
-			l_buf[0] = PPAT(-1);
-			r_buf[0] = PAT(1);
-			if (strcoll(l_buf, ch_buf) <= 0 &&
-			    strcoll(ch_buf, r_buf) <= 0)
-#else
-			if (PPAT(-1) <= ch && PAT(1) >= ch)
-#endif
-			    break;
-		    } else if (ch == PAT(0))
-			break;
+		rangematch(&pat, ch, Inbrack);
 		DPUTS(!pat || !*pat, "BUG: something is very wrong in doesmatch()");
 		if (*pat == Outbrack)
 		    break;
@@ -2458,20 +2533,6 @@ tokenize(char *s)
 	    }
 	    bslash = 1;
 	    continue;
-	case '[':
-	    if (bslash) {
-		s[-1] = Bnull;
-		break;
-	    }
-	    t = s;
-	    if (*++s == '^' || *s == '!')
-		s++;
-	    while (*s && *++s != ']');
-	    if (!*s)
-		return;
-	    *t = Inbrack;
-	    *s = Outbrack;
-	    break;
 	case '<':
 	    if (isset(SHGLOB))
 		break;
@@ -2499,6 +2560,8 @@ tokenize(char *s)
 	case ')':
 	    if (isset(SHGLOB))
 		break;
+	case '[':
+	case ']':
 	case '*':
 	case '?':
 	    for (t = ztokens; *t; t++)
