@@ -655,6 +655,12 @@ printcmdnamnode(HashNode hn, int printflags)
 {
     Cmdnam cn = (Cmdnam) hn;
 
+    if (printflags & PRINT_WHENCE_WORD) {
+	printf("%s: %s\n", cn->nam, (cn->flags & HASHED) ? 
+	       "hashed" : "command");
+	return;
+    }
+
     if ((printflags & PRINT_WHENCE_CSH) || (printflags & PRINT_WHENCE_SIMPLE)) {
 	if (cn->flags & HASHED) {
 	    zputs(cn->u.cmd, stdout);
@@ -813,10 +819,11 @@ printshfuncnode(HashNode hn, int printflags)
 	return;
     }
  
-    if ((printflags & PRINT_WHENCE_VERBOSE) &&
+    if ((printflags & (PRINT_WHENCE_VERBOSE|PRINT_WHENCE_WORD)) &&
 	!(printflags & PRINT_WHENCE_FUNCDEF)) {
 	nicezputs(f->nam, stdout);
-	printf(" is a shell function\n");
+	printf((printflags & PRINT_WHENCE_WORD) ? ": function\n" :
+	       " is a shell function\n");
 	return;
     }
  
@@ -911,6 +918,11 @@ printreswdnode(HashNode hn, int printflags)
 {
     Reswd rw = (Reswd) hn;
 
+    if (printflags & PRINT_WHENCE_WORD) {
+	printf("%s: reserved\n", rw->nam);
+	return;
+    }
+
     if (printflags & PRINT_WHENCE_CSH) {
 	printf("%s: shell reserved word\n", rw->nam);
 	return;
@@ -999,6 +1011,11 @@ printaliasnode(HashNode hn, int printflags)
 	return;
     }
 
+    if (printflags & PRINT_WHENCE_WORD) {
+	printf("%s: alias\n", a->nam);
+	return;
+    }
+
     if (printflags & PRINT_WHENCE_SIMPLE) {
 	zputs(a->text, stdout);
 	putchar('\n');
@@ -1187,6 +1204,7 @@ emptynameddirtable(HashTable ht)
 static void
 fillnameddirtable(HashTable ht)
 {
+#ifdef HAVE_GETPWENT
     if (!allusersadded) {
 	struct passwd *pw;
  
@@ -1201,6 +1219,7 @@ fillnameddirtable(HashTable ht)
 	allusersadded = 1;
     }
     return;
+#endif /* HAVE_GETPWENT */
 }
 
 /* Add an entry to the named directory hash *
