@@ -251,6 +251,14 @@ safeinungetc(int c)
 	inungetc(c);
 }
 
+/**/
+void
+herrflush(void)
+{
+    while (!lexstop && inbufct)
+	hwaddc(ingetc());
+}
+
 /* extract :s/foo/bar/ delimiters and arguments */
 
 /**/
@@ -272,8 +280,8 @@ getsubsargs(char *subline)
     zsfree(hsubr);
     hsubr = ptr2;
     if (hsubl && !strstr(subline, hsubl)) {
+	herrflush();
 	zerr("substitution failed", NULL, 0);
-	inerrflush();
 	return 1;
     }
     return 0;
@@ -358,7 +366,7 @@ histsubchar(int c)
 	    mev = ev = hconsearch(hsubl = ztrdup(buf), &marg);
 	    evset = 0;
 	    if (ev == -1) {
-		inerrflush();
+		herrflush();
 		zerr("no such event: %s", buf, 0);
 		return -1;
 	    }
@@ -410,9 +418,8 @@ histsubchar(int c)
 		ev = curhist;
 		evset = 1;
 	    } else if ((ev = hcomsearch(buf)) == -1) {
+		herrflush();
 		zerr("event not found: %s", buf, 0);
-		while (c != '\n' && !lexstop)
-		    c = ingetc();
 		return -1;
 	    } else
 		evset = 1;
@@ -434,9 +441,8 @@ histsubchar(int c)
 		    ehist = gethist(defev = mev);
 		    argc = getargc(ehist);
 		} else {
+		    herrflush();
 		    zerr("Ambiguous history reference", NULL, 0);
-		    while (c != '\n' && !lexstop)
-			c = ingetc();
 		    return -1;
 		}
 
@@ -493,28 +499,28 @@ histsubchar(int c)
 		break;
 	    case 'h':
 		if (!remtpath(&sline)) {
-		    inerrflush();
+		    herrflush();
 		    zerr("modifier failed: h", NULL, 0);
 		    return -1;
 		}
 		break;
 	    case 'e':
 		if (!rembutext(&sline)) {
-		    inerrflush();
+		    herrflush();
 		    zerr("modifier failed: e", NULL, 0);
 		    return -1;
 		}
 		break;
 	    case 'r':
 		if (!remtext(&sline)) {
-		    inerrflush();
+		    herrflush();
 		    zerr("modifier failed: r", NULL, 0);
 		    return -1;
 		}
 		break;
 	    case 't':
 		if (!remlpaths(&sline)) {
-		    inerrflush();
+		    herrflush();
 		    zerr("modifier failed: t", NULL, 0);
 		    return -1;
 		}
@@ -526,7 +532,7 @@ histsubchar(int c)
 		if (hsubl && hsubr)
 		    subst(&sline, hsubl, hsubr, gbal);
 		else {
-		    inerrflush();
+		    herrflush();
 		    zerr("no previous substitution", NULL, 0);
 		    return -1;
 		}
@@ -544,7 +550,7 @@ histsubchar(int c)
 		upcase(&sline);
 		break;
 	    default:
-		inerrflush();
+		herrflush();
 		zerr("illegal modifier: %c", NULL, c);
 		return -1;
 	    }
@@ -995,12 +1001,12 @@ getargspec(int argc, int marg, int evset)
 	ret = argc;
     else if (c == '%') {
 	if (evset) {
-	    inerrflush();
+	    herrflush();
 	    zerr("Ambiguous history reference", NULL, 0);
 	    return -2;
 	}
 	if (marg == -1) {
-	    inerrflush();
+	    herrflush();
 	    zerr("%% with no previous word matched", NULL, 0);
 	    return -2;
 	}
@@ -1233,7 +1239,7 @@ gethist(int ev)
 
     ret = quietgethist(ev);
     if (!ret) {
-	inerrflush();
+	herrflush();
 	zerr("no such event: %d", NULL, ev);
     }
     return ret;
@@ -1248,7 +1254,7 @@ getargs(Histent elist, int arg1, int arg2)
 
     if (arg2 < arg1 || arg1 >= nwords || arg2 >= nwords) {
 	/* remember, argN is indexed from 0, nwords is total no. of words */
-	inerrflush();
+	herrflush();
 	zerr("no such word in event", NULL, 0);
 	return NULL;
     }
