@@ -88,14 +88,19 @@ loop(int toplevel, int justonce)
     for (;;) {
 	freeheap();
 	errflag = 0;
-	if (interact && isset(SHINSTDIN))
-	    preprompt();
+	if (isset(SHINSTDIN)) {
+	    setblock_stdin();
+	    if (interact)
+		preprompt();
+	}
 	hbegin();		/* init history mech        */
 	intr();			/* interrupts on            */
 	lexinit();              /* initialize lexical state */
 	if (!(list = parse_event())) {	/* if we couldn't parse a list */
 	    hend();
-	    if ((tok == ENDINPUT && !errflag) || justonce)
+	    if ((tok == ENDINPUT && !errflag) ||
+		(tok == LEXERR && (!isset(SHINSTDIN) || !toplevel)) ||
+		justonce)
 		break;
 	    continue;
 	}
@@ -462,7 +467,9 @@ init_term(void)
 void
 setupvals(void)
 {
+#ifdef HAVE_GETPWUID
     struct passwd *pswd;
+#endif
     struct timezone dummy_tz;
     char *ptr;
 #ifdef HAVE_GETRLIMIT
@@ -552,10 +559,13 @@ setupvals(void)
     cached_uid = getuid();
 
     /* Get password entry and set info for `HOME' and `USERNAME' */
+#ifdef HAVE_GETPWUID
     if ((pswd = getpwuid(cached_uid))) {
 	home = metafy(pswd->pw_dir, -1, META_DUP);
 	cached_username = ztrdup(pswd->pw_name);
-    } else {
+    } else
+#endif /* HAVE_GETPWUID */
+	   {
 	home = ztrdup("/");
 	cached_username = ztrdup("");
     }
