@@ -101,8 +101,6 @@ static char *inbufpush;		/* Character at which to re-push alias */
 static int inbufleft;		/* Characters left in current input
 				   stack element */
 
-static int lastc;		/* used as flag that end of line was reached */
-
 
  /* Input must be stacked since the input queue is used by
   * various different parts of the shell.
@@ -178,8 +176,10 @@ shingetline(void)
 int
 ingetc(void)
 {
+    char lastc;
+
     if (lexstop)
-	return lastc = ' ';
+	return ' ';
     for (;;) {
 	if (inbufleft) {
 	    inbufleft--;
@@ -202,11 +202,11 @@ ingetc(void)
 	 */
 	if (strin || errflag) {
 	    lexstop = 1;
-	    return lastc = ' ';
+	    return ' ';
 	}
 	/* As a last resort, get some more input */
 	if (inputline())
-	    return lastc = ' ';
+	    return ' ';
     }
 }
 
@@ -422,12 +422,7 @@ stuff(char *fn)
 void
 inerrflush(void)
 {
-    /*
-     * This always goes character by character, but at present
-     * it is only used in the history code, where that is the only
-     * completely safe way of discarding input.
-     */
-    while ((strin || lastc != '\n') && !lexstop)
+    while (!lexstop && inbufct)
 	ingetc();
 }
 
