@@ -658,6 +658,13 @@ addproc(pid_t pid, char *text)
 	/* first process for this job */
 	jobtab[thisjob].procs = pn;
     }
+    /* If the first process in the job finished before any others were *
+     * added, maybe STAT_DONE got set incorrectly.  This can happen if *
+     * a $(...) was waited for and the last existing job in the        *
+     * pipeline was already finished.  We need to be very careful that *
+     * there was no call to printjob() between then and now, else      *
+     * the job will already have been deleted from the table.          */
+    jobtab[thisjob].stat &= ~STAT_DONE;
 }
 
 /* Check if we have files to delete.  We need to check this to see *
