@@ -876,27 +876,15 @@ gettokstr(int c, int sub)
 	    }
 	    break;
 	case LX2_INBRACK:
-	    add(c);
 	    if (!in_brace_param)
 		brct++;
-	    c = hgetc();
-	    if (c == '!' || c == '^') {
-		add(c);
-		c = hgetc();
-	    }
-	    if (c == ']')
-		break;
-	    if (lexstop)
-		goto brk;
-	    intpos = 0;
-	    continue;
+	    c = Inbrack;
+	    break;
 	case LX2_OUTBRACK:
 	    if (!in_brace_param)
 		brct--;
-	    if (brct < 0) {
+	    if (brct < 0)
 		brct = 0;
-		break;
-	    }
 	    c = Outbrack;
 	    break;
 	case LX2_INPAR:
