@@ -156,11 +156,12 @@ execselect(Cmd cmd)
     inp = fdopen(dup((SHTTY == -1) ? 0 : SHTTY), "r");
     selectlist(args);
     for (;;) {
-	do {
+	for (;;) {
 	    if (empty(bufstack)) {
-	    	if (interact && SHTTY != -1 && isset(USEZLE))
+	    	if (interact && SHTTY != -1 && isset(USEZLE)) {
+		    isfirstln = 1;
 		    str = (char *)zleread(prompt3, NULL, 0);
-	    	else {
+	    	} else {
 		    str = promptexpand(prompt3, 0, NULL, NULL);
 		    zputs(str, stderr);
 		    free(str);
@@ -178,11 +179,10 @@ execselect(Cmd cmd)
 	    }
 	    if ((s = strchr(str, '\n')))
 		*s = '\0';
-	    if(!*str) {
-		selectlist(args);
-		continue;
-	    }
-	} while(0);
+	    if (*str)
+	      break;
+	    selectlist(args);
+	}
 	setsparam("REPLY", ztrdup(str));
 	i = atoi(str);
 	if (!i)
