@@ -75,7 +75,11 @@ main(int argc, char **argv)
     for (;;) {
 	do
 	    loop(1,0);
-	while (tok != ENDINPUT);
+	while (tok != ENDINPUT && (tok != LEXERR || isset(SHINSTDIN)));
+	if (tok == LEXERR) {
+	    stopmsg = 1;
+	    zexit(lastval, 0);
+	}
 	if (!(isset(IGNOREEOF) && interact)) {
 #if 0
 	    if (interact)
