@@ -178,11 +178,7 @@ IPDEF5("LINES", &lines, zlevarsetfn),
 IPDEF5("OPTIND", &zoptind, intvarsetfn),
 IPDEF5("SHLVL", &shlvl, intvarsetfn),
 
-#define IPDEF6(A,B) {NULL,A,PM_SCALAR|PM_READONLY|PM_SPECIAL,BR((void *)B),SFN(nullsetfn),GFN(strvargetfn),stdunsetfn,0,NULL,NULL,NULL,0}
-IPDEF6("PWD", &pwd),
-
 #define IPDEF7(A,B) {NULL,A,PM_SCALAR|PM_SPECIAL,BR((void *)B),SFN(strvarsetfn),GFN(strvargetfn),stdunsetfn,0,NULL,NULL,NULL,0}
-IPDEF7("OLDPWD", &oldpwd),
 IPDEF7("OPTARG", &zoptarg),
 IPDEF7("NULLCMD", &nullcmd),
 IPDEF7("POSTEDIT", &postedit),
@@ -305,7 +301,7 @@ createparamtable(void)
 	setsparam("FCEDIT", ztrdup(DEFAULT_FCEDIT));
 	setsparam("TMPPREFIX", ztrdup(DEFAULT_TMPPREFIX));
 	setsparam("TIMEFMT", ztrdup(DEFAULT_TIMEFMT));
-	setsparam("WATCHFMT", ztrdup(DEFAULT_WATCHFMT));
+	setsparam("WATCHFMT", ztrdup(default_watchfmt));
 	setsparam("HOST", ztrdup(hostnam));
 	setsparam("LOGNAME", ztrdup((str = getlogin()) && *str ? str : cached_username));
 
@@ -346,11 +342,6 @@ createparamtable(void)
 	    pm->flags |= PM_EXPORTED;
 	    pm->env = addenv("HOME", home);
 	}
-	pm = (Param) paramtab->getnode(paramtab, "PWD");
-	if (!(pm->flags & PM_EXPORTED)) {
-	    pm->flags |= PM_EXPORTED;
-	    pm->env = addenv("PWD", pwd);
-	}
 	pm = (Param) paramtab->getnode(paramtab, "LOGNAME");
 	if (!(pm->flags & PM_EXPORTED)) {
 	    pm->flags |= PM_EXPORTED;
@@ -363,6 +354,7 @@ createparamtable(void)
 	pm->env = addenv("SHLVL", buf);
 
 	/* Add the standard non-special parameters */
+	set_pwd_env();
 	setsparam("MACHTYPE", ztrdup(MACHTYPE));
 	setsparam("OSTYPE", ztrdup(OSTYPE));
 	setsparam("TTY", ztrdup(ttystrname));
@@ -1339,7 +1331,14 @@ unsetparam_pm(Param pm, int altflag, int exp)
 	    unsetparam_pm(altpm, 1, exp);
     }
 
-    if (locallevel >= pm->level && (pm->flags & PM_SPECIAL))
+    /* If this was a local variable, we need to keep the old     *
+     * struct so that it is resurrected at the right level.      *
+     * This is partly because when an array/scalar value is set  *
+     * and the parameter used to be the other sort, unsetparam() *
+     * is called.  Beyond that, there is an ambiguity:  should   *
+     * foo() { local bar; unset bar; } make the global bar       *
+     * available or not?  The following makes the answer "no".   */
+    if (locallevel >= pm->level)
 	return;
 
     paramtab->removenode(paramtab, pm->nam); /* remove parameter node from table */
@@ -1652,20 +1651,20 @@ usernamegetfn(Param pm)
 void
 usernamesetfn(Param pm, char *x)
 {
-#ifdef HAVE_SETUID
+#if defined(HAVE_SETUID) && defined(HAVE_GETPWNAM)
     struct passwd *pswd;
 
     if (x && (pswd = getpwnam(x)) && (pswd->pw_uid != cached_uid)) {
-#ifdef HAVE_INITGROUPS
+# ifdef HAVE_INITGROUPS
 	initgroups(x, pswd->pw_gid);
-#endif
+# endif
 	if(!setgid(pswd->pw_gid) && !setuid(pswd->pw_uid)) {
 	    zsfree(cached_username);
 	    cached_username = ztrdup(pswd->pw_name);
 	    cached_uid = pswd->pw_uid;
 	}
     }
-#endif
+#endif /* HAVE_SETUID && HAVE_GETPWNAM */
 }
 
 /* Function to get value for special parameter `UID' */
