@@ -72,7 +72,7 @@ struct list dummy_list;
 
 #define YYERROR  { tok = LEXERR; return NULL; }
 #define YYERRORV { tok = LEXERR; return; }
-#define COND_ERROR(X,Y) do{zerr(X,Y,0);discard_input();YYERROR}while(0)
+#define COND_ERROR(X,Y) do{herrflush();zerr(X,Y,0);YYERROR}while(0)
 
 #define make_list()     allocnode(N_LIST)
 #define make_sublist()  allocnode(N_SUBLIST)
@@ -142,8 +142,8 @@ par_event(void)
 	    yyerror();
 	    return NULL;
 	}
+	herrflush();
 	yyerror();
-	discard_input();
 	return NULL;
     } else {
 	l->right = par_event();
@@ -542,16 +542,12 @@ par_case(Cmd c)
 
 	while (tok == SEPER)
 	    yylex();
-	if (tok == OUTBRACE) {
-	    yylex();
+	if (tok == OUTBRACE)
 	    break;
-	}
 	if (tok != STRING)
 	    YYERRORV;
-	if (!strcmp(tokstr, "esac")) {
-	    yylex();
+	if (!strcmp(tokstr, "esac"))
 	    break;
-	}
 	str = ncalloc(strlen(tokstr) + 2);
 	*str = ';';
 	strcpy(str + 1, tokstr);
@@ -618,10 +614,8 @@ par_case(Cmd c)
 	addlinknode(pats, str);
 	addlinknode(lists, par_list());
 	n++;
-	if ((tok == ESAC && !brflag) || (tok == OUTBRACE && brflag)) {
-	    yylex();
+	if ((tok == ESAC && !brflag) || (tok == OUTBRACE && brflag))
 	    break;
-	}
 	if(tok == SEMIAMP)
 	    *str = '&';
 	else if (tok != DSEMI)
@@ -631,6 +625,9 @@ par_case(Cmd c)
 	yylex();
     }
 
+    incmdpos = 1;
+    yylex();
+
     cc->pats = (char **)alloc((n + 1) * sizeof(char *));
 
     for (pp = cc->pats, no = firstnode(pats); no; incnode(no))
@@ -1364,24 +1361,6 @@ par_cond_triple(char *a, char *b, char *c)
     return n;
 }
 
-/**/
-static void
-discard_input(void)
-{
-    errflag = 0;
-    if (isnewlin <= 0) {
-	/* Discard remaining stuff after a parse error. */
-	int c;
-
-	hwbegin(0);
-	while ((c = hgetc()) != '\n' && !lexstop);
-	if (c == '\n')
-	    hungetc('\n');
-	hwend();
-    }
-    errflag = 1;
-}
-
 /**/
 static void
 yyerror(void)
