@@ -97,7 +97,7 @@ extern char *strstr _((const char *s, const char *t));
 #endif
 
 #ifndef HAVE_GETHOSTNAME
-extern int gethostname _((char *name, int namelen));
+extern int gethostname _((char *name, size_t namelen));
 #endif
 
 #ifndef HAVE_GETTIMEOFDAY
