@@ -83,29 +83,41 @@ char *alloca _((size_t));
 #endif
 
 #include <stdio.h>
-#include <pwd.h>
-#include <grp.h>
 #include <ctype.h>
 #include <sys/stat.h>
 #include <signal.h>
 #include <setjmp.h>
 
-#if HAVE_DIRENT_H
+#ifdef HAVE_PWD_H
+# include <pwd.h>
+#endif
+
+#ifdef HAVE_GRP_H
+# include <grp.h>
+#endif
+
+#ifdef HAVE_DIRENT_H
 # include <dirent.h>
-# define NLENGTH(dirent) strlen((dirent)->d_name)
-#else
-# define dirent direct
-# define NLENGTH(dirent) (dirent)->d_namlen
-# if HAVE_SYS_NDIR_H
+#else /* !HAVE_DIRENT_H */
+# ifdef HAVE_SYS_NDIR_H
 #  include <sys/ndir.h>
 # endif
-# if HAVE_SYS_DIR_H
+# ifdef HAVE_SYS_DIR_H
 #  include <sys/dir.h>
 # endif
-# if HAVE_NDIR_H
+# ifdef HAVE_NDIR_H
 #  include <ndir.h>
 # endif
-#endif
+# define dirent direct
+# undef HAVE_STRUCT_DIRENT_D_INO
+# undef HAVE_STRUCT_DIRENT_D_STAT
+# ifdef HAVE_STRUCT_DIRECT_D_INO
+#  define HAVE_STRUCT_DIRENT_D_INO HAVE_STRUCT_DIRECT_D_INO
+# endif
+# ifdef HAVE_STRUCT_DIRECT_D_STAT
+#  define HAVE_STRUCT_DIRENT_D_STAT HAVE_STRUCT_DIRECT_D_STAT
+# endif
+#endif /* !HAVE_DIRENT_H */
 
 #ifdef HAVE_STDLIB_H
 # ifdef ZSH_MEM
@@ -139,6 +151,12 @@ struct timezone {
 };
 #endif
 
+/* There's more than one non-standard way to get at this data */
+#if !defined(HAVE_STRUCT_DIRENT_D_INO) && defined(HAVE_STRUCT_DIRENT_D_STAT)
+# define d_ino d_stat.st_ino
+# define HAVE_STRUCT_DIRENT_D_INO HAVE_STRUCT_DIRENT_D_STAT
+#endif /* !HAVE_STRUCT_DIRENT_D_INO && HAVE_STRUCT_DIRENT_D_STAT */
+
 /* Sco needs the following include for struct utimbuf *
  * which is strange considering we do not use that    *
  * anywhere in the code                               */
@@ -298,12 +316,6 @@ struct timezone {
 # include <sys/utsname.h>
 #endif
 
-#if defined(HAVE_UTMPX_UT_HOST) || defined(HAVE_UTMP_UT_HOST)
-# define DEFAULT_WATCHFMT "%n has %a %l from %m."
-#else
-# define DEFAULT_WATCHFMT "%n has %a %l."
-#endif
-
 #define DEFAULT_WORDCHARS "*?_-.[]~=/&;!#$%^(){}<>"
 #define DEFAULT_TIMEFMT   "%J  %U user %S system %P cpu %*E total"
 
