@@ -413,6 +413,7 @@ char *cached_username;
 char *
 get_username(void)
 {
+#ifdef HAVE_GETPWUID
     struct passwd *pswd;
     uid_t current_uid;
  
@@ -425,6 +426,9 @@ get_username(void)
 	else
 	    cached_username = ztrdup("");
     }
+#else /* !HAVE_GETPWUID */
+    cached_uid = current_uid;
+#endif /* !HAVE_GETPWUID */
     return cached_username;
 }
 
@@ -541,7 +545,6 @@ getnameddir(char *name)
 {
     Param pm;
     char *str;
-    struct passwd *pw;
     Nameddir nd;
 
     /* Check if it is already in the named directory table */
@@ -558,14 +561,19 @@ getnameddir(char *name)
 	return str;
     }
 
-    /* Retrieve an entry from the password table/database for this user. */
-    if ((pw = getpwnam(name))) {
-	char *dir = xsymlink(pw->pw_dir);
-	adduserdir(name, dir, ND_USERNAME, 1);
-	str = dupstring(dir);
-	zsfree(dir);
-	return str;
+#ifdef HAVE_GETPWNAM
+    {
+	/* Retrieve an entry from the password table/database for this user. */
+	struct passwd *pw;
+	if ((pw = getpwnam(name))) {
+	    char *dir = xsymlink(pw->pw_dir);
+	    adduserdir(name, dir, ND_USERNAME, 1);
+	    str = dupstring(dir);
+	    zsfree(dir);
+	    return str;
+	}
     }
+#endif /* HAVE_GETPWNAM */
 
     /* There are no more possible sources of directory names, so give up. */
     return NULL;
@@ -1074,6 +1082,40 @@ zstrtol(const char *s, char **t, int base)
     return neg ? -ret : ret;
 }
 
+/**/
+int
+setblock_stdin(void)
+{
+#ifdef O_NDELAY
+# ifdef O_NONBLOCK
+#  define NONBLOCK (O_NDELAY|O_NONBLOCK)
+# else /* !O_NONBLOCK */
+#  define NONBLOCK O_NDELAY
+# endif /* !O_NONBLOCK */
+#else /* !O_NDELAY */
+# ifdef O_NONBLOCK
+#  define NONBLOCK O_NONBLOCK
+# else /* !O_NONBLOCK */
+#  define NONBLOCK 0
+# endif /* !O_NONBLOCK */
+#endif /* !O_NDELAY */
+
+#if NONBLOCK
+    struct stat st;
+    long mode;
+
+    if (!fstat(0, &st) && !S_ISREG(st.st_mode)) {
+	mode = fcntl(0, F_GETFL);
+	if (mode != -1 && (mode & NONBLOCK) &&
+	    !fcntl(0, F_SETFL, mode & ~NONBLOCK))
+	    return 1;
+    }
+#endif /* NONBLOCK */
+    return 0;
+
+#undef NONBLOCK
+}
+
 /**/
 int
 checkrmall(char *s)
@@ -1370,7 +1412,7 @@ ztrftime(char *buf, int bufsize, char *fmt, struct tm *tm)
 		*buf++ = '0' + tm->tm_sec % 10;
 		break;
 	    case 'y':
-		*buf++ = '0' + tm->tm_year / 10;
+		*buf++ = '0' + (tm->tm_year / 10) % 10;
 		*buf++ = '0' + tm->tm_year % 10;
 		break;
 #ifndef HAVE_STRFTIME
@@ -3617,7 +3659,7 @@ privasserted(void)
 {
     if(!geteuid())
 	return 1;
-#ifdef HAVE_CAP_INIT
+#ifdef HAVE_CAP_GET_PROC
     {
 	cap_t caps = cap_get_proc();
 	if(caps) {
@@ -3634,7 +3676,7 @@ privasserted(void)
 	    cap_free(&caps);
 	}
     }
-#endif /* HAVE_CAP_INIT */
+#endif /* HAVE_CAP_GET_PROC */
     return 0;
 }
 
