@@ -107,10 +107,28 @@
 #undef HAVE_STRUCT_UTMPX
 
 /* Define if your system's struct utmp has a member named ut_host.  */
-#undef HAVE_UTMP_UT_HOST
+#undef HAVE_STRUCT_UTMP_UT_HOST
 
 /* Define if your system's struct utmpx has a member named ut_host.  */
-#undef HAVE_UTMPX_UT_HOST
+#undef HAVE_STRUCT_UTMPX_UT_HOST
+
+/* Define if your system's struct utmpx has a member named ut_xtime.  */
+#undef HAVE_STRUCT_UTMPX_UT_XTIME
+
+/* Define if your system's struct utmpx has a member named ut_tv.  */
+#undef HAVE_STRUCT_UTMPX_UT_TV
+
+/* Define if your system's struct dirent has a member named d_ino.  */
+#undef HAVE_STRUCT_DIRENT_D_INO
+
+/* Define if your system's struct dirent has a member named d_stat.  */
+#undef HAVE_STRUCT_DIRENT_D_STAT
+
+/* Define if your system's struct direct has a member named d_ino.  */
+#undef HAVE_STRUCT_DIRECT_D_INO
+
+/* Define if your system's struct direct has a member named d_stat.  */
+#undef HAVE_STRUCT_DIRECT_D_STAT
 
 /* Define to be a string corresponding the vendor of the machine */
 #undef VENDOR
